import logging

from sos_access.exceptions import (AlarmReceiverConnectionError,
                                   TCPTransportError, NotTreatedNotDistributed,
                                   OtherError)

logger = logging.getLogger(__name__)

fail_over_errors = (TCPTransportError, NotTreatedNotDistributed, OtherError)


def alternating_retry(func):
    """
    Decorator function that will allow for retrying alternately between the
    primary and secondary alarm receiver server.
    """

    def retried_func(*args, **kwargs):
        use_secondary = kwargs.get('secondary', False)
        retry_count = 0
        client = args[0]

        # if we only have a single receiver we only have to try on that.
        if client.use_single_receiver:
            max_retries = client.MAX_RETRY
        else:
            max_retries = client.MAX_RETRY * 2

        while retry_count < max_retries:
            try:
                kwargs['secondary'] = use_secondary
                result = func(*args, **kwargs)
                return result
            except fail_over_errors as e:
                logger.exception(e)
                if not client.use_single_receiver:
                    use_secondary = not use_secondary  # toggle fail over
                retry_count = retry_count + 1

        # if it comes out of the loop we raise new exeption.
        raise AlarmReceiverConnectionError('Not possible to send data to any '
                                           'of the client receivers')

    return retried_func



import marshmallow
from marshmallow.validate import Length, OneOf
import xmltodict

ALLOWED_STATUS_CODES = [0, 1, 2, 3, 4, 5, 7, 9, 10, 98, 99, 100, 101]

ALARM_TYPES = ['AL', 'RE']


# TODO: write tests for all schemas.


class SOSAccessRequest:
    pass


class AlarmRequest(SOSAccessRequest):

    def __init__(self, event_code, transmitter_type, transmitter_code,
                 authentication, receiver, alarm_type=None,
                 transmitter_time=None, reference=None, transmitter_area=None,
                 section=None, section_text=None, detector=None,
                 detector_text=None, additional_info=None, position=None):
        self.event_code = event_code
        self.transmitter_type = transmitter_type
        self.transmitter_code = transmitter_code
        self.authentication = authentication
        self.receiver = receiver
        self.alarm_type = alarm_type
        self.transmitter_time = transmitter_time
        self.reference = reference
        self.transmitter_area = transmitter_area
        self.section = section
        self.section_text = section_text
        self.detector = detector
        self.detector_text = detector_text
        self.additional_info = additional_info
        self.position = position

    @property
    def additional_info_text(self):
        """
        Additional info is extra info about the alarm. Input is added using the
        additional_info but we are under constraint to format the output
        properly. We accept a list of values. Each item should be printed on a
        separate line. We use CR+LF to separate lines.
        If we dont get a list we try to make a string of it and send it on.
        We must keep the resulting text under 2000 chars. On list input we
        truncate the last input that makes the message go over the limit
        On single row we truncate the data and add ... to it.

        :return: text
        """
        text = ''
        max_length = 2000
        if isinstance(self.additional_info, (list, tuple)):

            for item in self.additional_info:
                line = f'{item} \r\n'
                if len(text + line) > max_length:
                    break
                text = text + line  # TODO: Test
            return text

        elif isinstance(self.additional_info, dict):
            # print key and value on each row.
            for key, value in self.additional_info.items():
                line = f'{key}: {value} \r\n'
                if len(text + line) > max_length:
                    break
                text = text + line  # TODO: Test

        else:
            line = str(self.additional_info)
            if len(line) > max_length:
                line = line[:1997]  # TODO: Test
            return line

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'event_code={self.event_code}, '
                f'transmitter_type={self.transmitter_type}, '
                f'transmitter_code={self.transmitter_code}), '
                f'receiver={self.receiver})')


class AlarmResponse(SOSAccessRequest):

    def __init__(self, status, info, arrival_time=None, reference=None):
        self.reference = reference
        self.status = status
        self.info = info
        self.arrival_time = arrival_time

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'status={self.status}, '
                f'info={self.info}, '
                f'arrival_time={self.arrival_time}, '
                f'reference={self.reference})')


class NewAuthRequest(SOSAccessRequest):

    def __init__(self, authentication, transmitter_code, transmitter_type,
                 reference=None):
        self.authentication = authentication
        self.reference = reference
        self.transmitter_code = transmitter_code
        self.transmitter_type = transmitter_type

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'authentication={self.authentication}, '
                f'transmitter_code={self.transmitter_code}, '
                f'transmitter_type={self.transmitter_type}, '
                f'reference={self.reference})')


class NewAuthResponse(SOSAccessRequest):

    def __init__(self, status, info, new_authentication, arrival_time=None,
                 reference=None):
        self.reference = reference
        self.status = status
        self.info = info
        self.new_authentication = new_authentication
        self.arrival_time = arrival_time

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'new_authentication=<redacted>, '
                f'status={self.status}, '
                f'info={self.info}, '
                f'arrival_time={self.arrival_time}, '
                f'reference={self.reference})')


class PingRequest(SOSAccessRequest):

    def __init__(self, authentication, transmitter_code, transmitter_type,
                 reference=None):
        self.authentication = authentication
        self.reference = reference
        self.transmitter_code = transmitter_code
        self.transmitter_type = transmitter_type

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'authentication=<redacted>, '
                f'transmitter_code={self.transmitter_code}, '
                f'transmitter_type={self.transmitter_type}, '
                f'reference={self.reference})')


class PingResponse(SOSAccessRequest):
    def __init__(self, status, info, arrival_time=None, reference=None):
        self.reference = reference
        self.status = status
        self.info = info
        self.arrival_time = arrival_time  # Alarm Request

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'status={self.status}, '
                f'info={self.info}, '
                f'arrival_time={self.arrival_time}, '
                f'reference={self.reference})')


class PositionSchema(marshmallow.Schema):
    """
    <pos> Geographical coordinate in the format
    RT90 (2,5 gon West):
    “xXXXXXXXyYYYYYYY”

    where x is the x-coordinate, y is the y-
    coordinate. Values are given in meters.

    Ex. ”x1234567y1234567”.

    or in the format WGS84 (Lat/Long):
    “NDDMMmmEDDDMMmm”
    where DD are degrees; MM minutes; mm
    decimal minutes (leading 0 shall be given on
    the longitude if needed).

    ex WGS84
    <position>
    <pos>E597295E0176288</pos>
    </position>
    Ex RT90
    <position>
    <pos>x1234567y1234567</pos>
    <position>
    """
    pos = marshmallow.fields.String(required=True,
                                    validate=[Length(min=14, max=16)])

    class Meta:
        ordered = True


class SOSAccessSchema(marshmallow.Schema):
    __envelope__ = None
    __model__ = None

    @marshmallow.pre_load()
    def load_xml(self, data):
        # incoming XML
        parsed_data = xmltodict.parse(data)
        # remove envelope
        return parsed_data[self.__envelope__]

    @marshmallow.post_dump()
    def dump_xml(self, data):
        # add the envelope
        data_to_dump = {self.__envelope__: data}
        # make xml
        return xmltodict.unparse(data_to_dump)

    @marshmallow.post_load()
    def make_object(self, data):
        return self.__model__(**data)


class AlarmRequestSchema(SOSAccessSchema):
    __envelope__ = 'alarmrequest'
    __model__ = AlarmRequest

    reference = marshmallow.fields.String(allow_none=True,
                                          validate=[Length(min=1, max=50)])
    authentication = marshmallow.fields.String(required=True,
                                               validate=[Length(equal=15)])
    receiver = marshmallow.fields.String(required=True,
                                         validate=[Length(min=1, max=20)])
    transmitter_time = marshmallow.fields.DateTime(allow_none=True,
                                                   data_key='transmittertime')
    alarm_type = marshmallow.fields.String(allow_none=True,
                                           validate=[OneOf(ALARM_TYPES)],
                                           data_key='alarmtype')
    transmitter_type = marshmallow.fields.String(required=True,
                                                 validate=[Length(equal=5)],
                                                 data_key='transmittertype')
    transmitter_code = marshmallow.fields.String(required=True, validate=[
        Length(min=1, max=15)], data_key='transmittercode')
    transmitter_area = marshmallow.fields.String(allow_none=True, validate=[
        Length(min=1, max=5)], data_key='transmitterarea')
    event_code = marshmallow.fields.String(required=True,
                                           validate=[Length(min=1, max=25)],
                                           data_key='eventcode')
    section = marshmallow.fields.String(allow_none=True,
                                        validate=[Length(min=1, max=5)])
    section_text = marshmallow.fields.String(allow_none=True,
                                             validate=[Length(min=1, max=40)],
                                             data_key='sectionkey')
    detector = marshmallow.fields.String(allow_none=True,
                                         validate=[Length(min=1, max=5)])
    detector_text = marshmallow.fields.String(allow_none=True,
                                              validate=[Length(min=1, max=40)],
                                              data_key='detectortext')
    # Lines in additionalinfo is separated via CR+LF or LF. CR = 0x0a LF = 0x0d
    additional_info = marshmallow.fields.String(allow_none=True, validate=[
        Length(min=1, max=2000)], data_key='additinalinfo', load_only=True)
    additional_info_text = marshmallow.fields.String(allow_none=True, validate=[
        Length(min=1, max=2000)], data_key='additinalinfo', dump_only=True)

    position = marshmallow.fields.Nested(PositionSchema, allow_none=True)

    class Meta:
        ordered = True


# Alarm Response


class AlarmResponseSchema(SOSAccessSchema):
    __envelope__ = 'alarmresponse'
    __model__ = AlarmResponse

    reference = marshmallow.fields.String(allow_none=True,
                                          validate=[Length(min=1, max=50)])
    status = marshmallow.fields.Integer(required=True,
                                        validate=[OneOf(ALLOWED_STATUS_CODES)])
    info = marshmallow.fields.String(required=True,
                                     validate=[Length(min=1, max=255)])
    arrival_time = marshmallow.fields.DateTime(allow_none=True,
                                               data_key='arrivaltime',
                                               format='rfc')

    class Meta:
        ordered = True


# Request new authentication

class NewAuthRequestSchema(SOSAccessSchema):
    __envelope__ = 'requestnewauthentication'
    __model__ = NewAuthRequest

    authentication = marshmallow.fields.String(required=True,
                                               validate=[Length(equal=15)])
    reference = marshmallow.fields.String(allow_none=True,
                                          validate=[Length(min=1, max=50)])
    transmitter_code = marshmallow.fields.String(required=True, validate=[
        Length(min=1, max=15)], data_key='transmittercode')
    transmitter_type = marshmallow.fields.String(required=True,
                                                 validate=[Length(equal=5)],
                                                 data_key='transmittertype')

    class Meta:
        ordered = True


# Request new authentication response


class NewAuthResponseSchema(SOSAccessSchema):
    __envelope__ = 'requestnewauthenticationresponse'
    __model__ = NewAuthResponse

    reference = marshmallow.fields.String(allow_none=True,
                                          validate=[Length(min=1, max=50)])
    status = marshmallow.fields.Integer(required=True,
                                        validate=[OneOf(ALLOWED_STATUS_CODES)])
    info = marshmallow.fields.String(required=True,
                                     validate=[Length(min=1, max=255)])
    new_authentication = marshmallow.fields.String(required=True,
                                                   validate=[Length(equal=15)],
                                                   data_key='newauthentication')
    arrival_time = marshmallow.fields.DateTime(allow_none=True,
                                               data_key='arrivaltime',
                                               format='rfc')

    class Meta:
        ordered = True


# Ping request


class PingRequestSchema(SOSAccessSchema):
    __envelope__ = 'pingrequest'
    __model__ = PingRequest

    authentication = marshmallow.fields.String(required=True,
                                               validate=[Length(equal=15)])
    reference = marshmallow.fields.String(allow_none=True,
                                          validate=[Length(min=1, max=50)])
    transmitter_code = marshmallow.fields.String(required=True, validate=[
        Length(min=1, max=15)], data_key='transmittercode')
    transmitter_type = marshmallow.fields.String(required=True,
                                                 validate=[Length(equal=5)],
                                                 data_key='transmittertype')

    class Meta:
        ordered = True


# Ping response


class PingResponseSchema(SOSAccessSchema):
    __envelope__ = 'pingresponse'
    __model__ = PingResponse

    reference = marshmallow.fields.String(allow_none=True,
                                          validate=[Length(min=1, max=50)])
    status = marshmallow.fields.Integer(required=True,
                                        validate=[OneOf(ALLOWED_STATUS_CODES)])
    info = marshmallow.fields.String(required=True,
                                     validate=[Length(min=1, max=255)])
    arrival_time = marshmallow.fields.DateTime(allow_none=True,
                                               data_key='arrivaltime',
                                               format='rfc')

    class Meta:
        ordered = True



class SOSAccessClientException(Exception):
    """General exception in client"""


class SOSAccessServerException(Exception):
    """General exception in server"""


class IncorrectlyConfigured(SOSAccessClientException):
    """The client is incorrectly configured"""


class SOSAccessError(SOSAccessClientException):
    """General Exception in transfer of alarms"""


class InvalidLength(SOSAccessError):
    """Message is too long"""


class InvalidXML(SOSAccessError):
    """Invalid XML content"""


class WrongContent(SOSAccessError):
    """Incorrect data in content. ex: too long text in a field."""


class NotAuthorized(SOSAccessError):
    """Not autorized. wrong combination of transmitter, instance or password."""


class NotTreatedNotDistributed(SOSAccessError):
    """Alarm not treated or distributed."""  # Use failover address


class MandatoryDataMissing(SOSAccessError):
    """Mandatory XML tag missing"""


class ServiceUnavailable(SOSAccessError):
    """Not autorized to use heartbeat service"""


class DuplicateAlarm(SOSAccessError):
    """Same alarm received multiple times"""


class OtherError(SOSAccessError):
    """Unknown receiver error"""  # use failover address.


class XMLHeaderError(SOSAccessError):
    """XML Header missing, or invalid"""


class PingToOften(SOSAccessError):
    """Heartbeat is sent too often"""


class ServerSystemError(SOSAccessServerException):
    """System error in receiving server"""


class TCPTransportError(Exception):
    """Something went wrong in sending data"""


class AlarmReceiverConnectionError(SOSAccessClientException):
    """Not possible to connect to the Alarm receivers specified"""



from sos_access.client import SOSAccessClient
from sos_access.schemas import AlarmRequest

"""
Krypterad:

Transmittercode: IK980102

Transmittertype: SV301

Authentication: 000000000000000

 

Okrypterad:

Transmittercode: IO980102

Transmittertype: SV300

Authentication: 000000000000000

 

Anslutningsuppgifter:

IP: 194.14.58.16/194.14.60.16

Krypterad port: 19100

Okrypterad port: 19000
"""

TRANSMITTER_CODE = 'IK980102'
TRANSMITTER_TYPE = 'SV301'
AUTH = '000000000000000'

ADDRESS = ('194.14.58.16', 19100)

client = SOSAccessClient(transmitter_code=TRANSMITTER_CODE,
                         transmitter_type=TRANSMITTER_TYPE, authentication=AUTH,
                         receiver_address=ADDRESS, receiver_id='SOSA',
                         use_single_receiver=True, use_tls=True)

client.send_alarm(event_code='AL')



from sos_access.client import SOSAccessClient
from sos_access.schemas import AlarmRequest

"""
Krypterad:

Transmittercode: IK980102

Transmittertype: SV301

Authentication: 000000000000000

 

Okrypterad:

Transmittercode: IO980102

Transmittertype: SV300

Authentication: 000000000000000

 

Anslutningsuppgifter:

IP: 194.14.58.16/194.14.60.16

Krypterad port: 19100

Okrypterad port: 19000
"""

TRANSMITTER_CODE = 'IO980102'
TRANSMITTER_TYPE = 'SV300'
AUTH = '*Tlz!3dLWSQuYsv'

ADDRESS = ('194.14.58.16', 19000)

# client = SOSAccessClient(transmitter_code=TRANSMITTER_CODE,
#                         transmitter_type=TRANSMITTER_TYPE, authentication=AUTH,
#                         receiver_address=ADDRESS, receiver_id='SOSA',
#                         use_single_receiver=True)

client = SOSAccessClient(transmitter_code=TRANSMITTER_CODE,
                         transmitter_type=TRANSMITTER_TYPE, authentication=AUTH,
                         receiver_address=ADDRESS, receiver_id='SOSA',
                         secondary_receiver_address=('194.14.58.16', 19000))

client.send_alarm(event_code='AL', additional_info={'Testing testing': 'cool',
                                                    'testing 2 testing 2': 'awesome'})
# client.restore_alarm(event_code='AL')

# client.ping()
# Not the password then testing request new password. Otherwise youll get locked out!
# client.request_new_auth()



import os
import sys
from shutil import rmtree
from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'sos-access'
DESCRIPTION = "Python client for sending alarm information via SOS Access v4"
URL = 'https://github.com/pwitab/sos-access'
EMAIL = 'henrik@pwit.se'
AUTHOR = "Henrik Palmlund Wahlgren @ Palmlund Wahlgren Innovative Technology AB"
REQUIRES_PYTHON = '>=3.6'
# VERSION = None

# What packages are required for this module to be executed?
REQUIRED = [
    'marshmallow>=3.0.0b1',
    'xmltodict']

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

here = os.path.abspath(os.path.dirname(__file__))


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system(
            '{0} setup.py sdist bdist_wheel'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        # os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

setup(
    name=NAME,
    version='0.0.4',
    python_requires=REQUIRES_PYTHON,
    description=DESCRIPTION,
    long_description=readme + '\n\n' + history,
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    entry_points={
    },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license="BSD-3",
    zip_safe=False,
    keywords=['alarm', 'sos access'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)



import logging
import socket
import ssl

from sos_access.exceptions import (IncorrectlyConfigured, InvalidLength,
                                   InvalidXML, WrongContent, NotAuthorized,
                                   NotTreatedNotDistributed,
                                   MandatoryDataMissing, ServiceUnavailable,
                                   DuplicateAlarm, OtherError, XMLHeaderError,
                                   PingToOften, ServerSystemError,
                                   TCPTransportError)
from sos_access.schemas import (AlarmRequestSchema, AlarmRequest,
                                AlarmResponseSchema, PingRequest,
                                PingRequestSchema, PingResponseSchema,
                                NewAuthRequestSchema, NewAuthResponseSchema,
                                NewAuthRequest, PingResponse, AlarmResponse,
                                NewAuthResponse)

from sos_access.decorators import alternating_retry

logger = logging.getLogger(__name__)


# TODO: Implement Point class for different ways of generation a geografical
#       point and use to position.
# TODO: Implement way of adding additionalText to an alarm. List of info?
# TODO: What event codes are there? An official list?
# TODO: Find better description on what Section and detector is used for in the recieiver
# TODO: Document monitored connections
# TODO: add proper logging
# TODO: convert time to utc aware


class SOSAccessClient:
    """
    Client implementing the SOS Access v4 protocol to be used for sending Alarm
    Requests to alarm operators.

    :param transmitter_code:
    :param transmitter_type:
    :param authentication:
    :param receiver_id:
    :param receiver_address:
    :param secondary_receiver_address:
    :param use_single_receiver:
    :param use_tls:
    """
    MAX_RETRY = 3
    ENCODING = 'latin-1'  # it is in the specs only to allow iso-8859-1

    def __init__(self, transmitter_code, transmitter_type, authentication,
                 receiver_id, receiver_address, secondary_receiver_address=None,
                 use_single_receiver=False, use_tls=False):
        self.transmitter_code = transmitter_code
        self.transmitter_type = transmitter_type
        self.authentication = authentication
        self.receiver_id = receiver_id
        self.receiver_address = receiver_address
        self.secondary_receiver_address = secondary_receiver_address
        self.use_single_receiver = use_single_receiver
        self.use_tls = use_tls

        if self.secondary_receiver_address is None and not use_single_receiver:
            raise IncorrectlyConfigured(
                'Both primary and secondary receiver address is needed.')

        # load all schemas so we don't have to remember to do it again.
        self.alarm_request_schema = AlarmRequestSchema()
        self.alarm_response_schema = AlarmResponseSchema()
        self.ping_request_schema = PingRequestSchema()
        self.ping_response_schema = PingResponseSchema()
        self.new_auth_request_schema = NewAuthRequestSchema()
        self.new_auth_response_schema = NewAuthResponseSchema()

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'transmitter_code={self.transmitter_code}, '
                f'transmitter_type={self.transmitter_type}, '
                f'authentication=<redacted>, '
                f'receiver_id={self.receiver_id}, '
                f'receiver_address={self.receiver_address}, '
                f'secondary_receiver_address={self.secondary_receiver_address}, '
                f'use_single_receiver={self.use_single_receiver}, '
                f'use_tls={self.use_tls})')

    def send_alarm(self, event_code, transmitter_time=None, reference=None,
                   transmitter_area=None, section=None, section_text=None,
                   detector=None, detector_text=None, additional_info=None,
                   position=None) -> AlarmResponse:
        """
        Sends an alarm in the receiver.

        :param event_code:
        :param transmitter_time:
        :param reference:
        :param transmitter_area:
        :param section:
        :param section_text:
        :param detector:
        :param detector_text:
        :param additional_info:
        :param position:
        :return:
        """
        alarm_request = AlarmRequest(event_code=event_code,
                                     transmitter_type=self.transmitter_type,
                                     transmitter_code=self.transmitter_code,
                                     authentication=self.authentication,
                                     receiver=self.receiver_id, alarm_type='AL',
                                     transmitter_time=transmitter_time,
                                     reference=reference,
                                     transmitter_area=transmitter_area,
                                     section=section, section_text=section_text,
                                     detector=detector,
                                     detector_text=detector_text,
                                     additional_info=additional_info,
                                     position=position)
        logger.info(f'Sending alarm request: {alarm_request}')

        return self._send_alarm(alarm_request)

    def restore_alarm(self, event_code, transmitter_time=None, reference=None,
                      transmitter_area=None, section=None, section_text=None,
                      detector=None, detector_text=None, additional_info=None,
                      position=None) -> AlarmResponse:
        """
        Restores an alarm in the receiver.

        :param event_code:
        :param transmitter_time:
        :param reference:
        :param transmitter_area:
        :param section:
        :param section_text:
        :param detector:
        :param detector_text:
        :param additional_info:
        :param position:
        :return:
        """
        alarm_request = AlarmRequest(event_code=event_code,
                                     transmitter_type=self.transmitter_type,
                                     transmitter_code=self.transmitter_code,
                                     authentication=self.authentication,
                                     receiver=self.receiver_id, alarm_type='RE',
                                     transmitter_time=transmitter_time,
                                     reference=reference,
                                     transmitter_area=transmitter_area,
                                     section=section, section_text=section_text,
                                     detector=detector,
                                     detector_text=detector_text,
                                     additional_info=additional_info,
                                     position=position)
        logger.info(f'Sending restore alarm request: {alarm_request}')

        return self._send_alarm(alarm_request)

    @alternating_retry
    def _send_alarm(self, alarm_request: AlarmRequest,
                    secondary=False) -> AlarmResponse:
        out_data = self.alarm_request_schema.dump(alarm_request)
        logger.debug(f'Sending SOS Access Data: {out_data}')
        in_data = self.transmit(out_data, secondary=secondary)
        logger.debug(f'Received SOS Access Data: {in_data}')
        alarm_response = self.alarm_response_schema.load(in_data)
        logger.info(f'Received alarm response: {alarm_response}')
        self._check_response_status(alarm_response)
        return alarm_response

    def ping(self, reference=None) -> PingResponse:
        """
        Sends a heart beat message to indicate to the alarm operator that
        the alarm device is still operational
        """

        ping_request = PingRequest(authentication=self.authentication,
                                   transmitter_code=self.transmitter_code,
                                   transmitter_type=self.transmitter_type,
                                   reference=reference)
        logger.info(f'Sending {ping_request}')
        return self._send_ping(ping_request)

    @alternating_retry
    def _send_ping(self, ping_request: PingRequest,
                   secondary=False) -> PingResponse:
        """
        To send ping request
        """
        out_data = self.ping_request_schema.dump(ping_request)
        logger.debug(f'Sending SOS Access data: {out_data}')
        in_data = self.transmit(out_data, secondary=secondary)
        logger.debug(f'Received SOS Access data: {in_data}')
        ping_response = self.ping_response_schema.load(in_data)
        logger.info(f'Received {ping_response}')
        self._check_response_status(ping_response)
        return ping_response

    def request_new_auth(self, reference=None) -> NewAuthResponse:
        """
        Send a request for new password on the server. This is used so that
        you can have a standard password when deploying devices but it is
        changed to something the alarm installer does not know when the alarm
        is operational
        """

        new_auth_request = NewAuthRequest(authentication=self.authentication,
                                          transmitter_code=self.transmitter_code,
                                          transmitter_type=self.transmitter_type,
                                          reference=reference)
        logger.info(f'Sending {new_auth_request}')
        return self._send_request_new_auth(new_auth_request)

    @alternating_retry
    def _send_request_new_auth(self, new_auth_request: NewAuthRequest,
                               secondary=False) -> NewAuthResponse:
        """
        To send request new auth request
        """
        out_data = self.new_auth_request_schema.dump(new_auth_request)
        logger.debug(f'Sending SOS Access data: {out_data}')
        in_data = self.transmit(out_data, secondary=secondary)
        logger.debug(f'Received SOS Access data: {in_data}')
        new_auth_response = self.new_auth_response_schema.load(in_data)
        logger.info(f'Received {new_auth_response}')
        self._check_response_status(new_auth_response)
        return new_auth_response

    def transmit(self, data, secondary=False):
        """
        Will create a TCP connection and send the request and received the
        response
        """
        if secondary:
            address = self.secondary_receiver_address
        else:
            address = self.receiver_address
        logger.info(f'Starting new connection to {address} with '
                    f'secure={self.use_tls}')

        with TCPTransport(address, secure=self.use_tls) as transport:
            transport.connect()
            transport.send(data.encode(self.ENCODING))
            response = transport.receive().decode(self.ENCODING)
            return response

    @staticmethod
    def _check_response_status(response):
        """
        Checks the status of the response and raise appropriate errors if the
        request wasn't successful.

        :param response:
        """
        # TODO: Test!
        if response.status == 1:
            raise InvalidLength(f'{response.info}: Message is too long.')
        elif response.status == 2:
            raise InvalidXML(f'{response.info}: Invalid XML content.')
        elif response.status == 3:
            raise WrongContent(f'{response.info}: Wrong data content. '
                               f'Ex: too long text in a field.')
        elif response.status == 4:
            raise NotAuthorized(f'{response.info}: Wrong combination of '
                                f'transmitter, instance and password')
        elif response.status == 5:
            # Fail over should kick in.
            raise NotTreatedNotDistributed(f'{response.info}: Error in '
                                           f'processing. It has neither been '
                                           f'treated or distributed')
        elif response.status == 7:
            raise MandatoryDataMissing(f'{response.info}: Mandatory XML tag '
                                       f'missing')
        elif response.status == 9:
            raise ServiceUnavailable(
                (f'{response.info}: Heartbeat is not enabled on the server for '
                 f'this transmitter or you are not authorized to use it.'))
        elif response.status == 10:
            raise DuplicateAlarm(f'{response.info}: The same alarm was received'
                                 f' multiple times')
        elif response.status == 98:
            raise ServerSystemError(f'{response.info}: General receiver error')
        elif response.status == 99:
            # Failover should kick in.
            raise OtherError(f'{response.info}: Unknown receiver error')
        elif response.status == 100:
            raise XMLHeaderError(f'{response.info}: Invalid or missing XML '
                                 f'header')
        elif response.status == 101:
            raise PingToOften(f'{response.info}: Heartbeat is sent too often')


class TCPTransport:

    def __init__(self, address, secure=False, timeout=5):
        self.address = address
        self.timeout = timeout
        self.socket = self._get_socket(secure, timeout)

    def __enter__(self):
        # If we dont do anything in __enter__ we dont have to handle exceptions
        # twice....
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.socket.close()
        # Catch any error related to the socket and raise as TCP error
        if exc_type in (
                OSError, IOError, ssl.SSLError, socket.error, socket.timeout):
            raise TCPTransportError from exc_type(exc_val, exc_tb)

    def _get_socket(self, secure=False, timeout=None):
        """
        Returns socket. If secure is True the socket is wrapped in an
        SSL Context
        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(timeout or self.timeout)

        if secure:
            # Setting Purpose to CLIENT_AUTH might seem a bit backwards. But
            # SOS Access v4 is using SSL/TLS for encryption not authentications
            # and verification. There is no cert and no hostname to check so
            # setting the purpose to Client Auth diables that in a nifty way.
            self.context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            return self.context.wrap_socket(sock)

        else:
            return sock

    def connect(self, address=None):
        _address = address or self.address
        self.socket.connect(_address)

    def send(self, data):
        """Send data over socket with correct encoding"""
        self.socket.sendall(data)

    def receive(self):
        """Receive data on socket and decode using correct encoding"""
        data = self.socket.recv(4096)
        return data
